@@ -1,8 +1,8 @@
 #! /usr/bin/env python
 # -*- coding: utf-8 -*-
 
-## Copyright 2000-2010 by LivingLogic AG, Bayreuth/Germany.
-## Copyright 2000-2010 by Walter Dörwald
+## Copyright 2000-2012 by LivingLogic AG, Bayreuth/Germany.
+## Copyright 2000-2012 by Walter Dörwald
 ##
 ## All Rights Reserved
 ##
@@ -14,9 +14,7 @@
 
 There will be no more than one sisyphus job of a certain name running at every
 given time. A job has a maximum allowed runtime. If this maximum is exceeded,
-the job will kill itself.
-
-In addition to that, job execution can be logged.
+the job will kill itself. In addition to that, job execution can be logged.
 
 To use this module, you must derive your own class from :class:`Job` and
 implement the :meth:`execute` method.
@@ -54,11 +52,11 @@ The following example illustrates the use of this module::
 			self.log("fetching data from {!r}".format(self.url))
 			data = urllib.urlopen(self.url).read()
 			datasize = len(data)
-			self.log("writing file {!r} ({} bytes)".format(self.tmpname, datasize))
+			self.log("writing file {!r} ({:,} bytes)".format(self.tmpname, datasize))
 			open(self.tmpname, "wb").write(data)
 			self.log("renaming file {!r} to {!r}".format(self.tmpname, self.officialname))
 			os.rename(self.tmpname, self.officialname)
-			return "cached {!r} as {!r} ({} bytes)".format(self.url, self.officialname, datasize)
+			return "cached {!r} as {!r} ({:,} bytes)".format(self.url, self.officialname, datasize)
 
 	if __name__=="__main__":
 		sisyphus.executewithargs(Fetch())
@@ -67,7 +65,7 @@ You will find the log files for this job in ``~/ll.sisyphus/ACME.FooBar/Fetch/``
 """
 
 
-import sys, os, signal, fcntl, codecs, traceback, errno, pprint, datetime, re, contextlib, argparse
+import sys, os, signal, fcntl, codecs, traceback, errno, pprint, datetime, re, contextlib, argparse, tokenize
 
 from ll import url, ul4c, misc
 
@@ -75,15 +73,6 @@ from ll import url, ul4c, misc
 __docformat__ = "reStructuredText"
 
 
-def literaldecode(exc):
-	return (u"".join(u"[%02x]" % ord(c) for c in exc.object[exc.start:exc.end]), exc.end)
-
-codecs.register_error("literaldecode", literaldecode)
-
-
-encodingdeclaration = re.compile(r"coding[:=]\s*([-\w.]+)")
-
-
 class MaximumRuntimeExceeded(Exception):
 	def __init__(self, maxtime):
 		self.maxtime = maxtime
@@ -95,10 +84,6 @@ class MaximumRuntimeExceeded(Exception):
 class Job(object):
 	"""
 	A Job object executes a task once.
-	
-	The job announces its presence (and its process id) in a file that is stored
-	in the :dir:`~/run/` directory. Logs will be created in the :dir:`~/log/`
-	directory (This can be changes by deriving new subclasses).
 
 	To use this class, derive your own class from it and overwrite the
 	:meth:`execute` method.
@@ -144,8 +129,8 @@ class Job(object):
 		hierarchical project name (e.g. including customer names or similar stuff).
 
 	``jobname`` : :option:`-j` or :option:`--jobname`
-		The name of the job itself (defaulting to the name of the class if
-		:const:`None` is given).
+		The name of the job itself (defaulting to the name of the class if none
+		is given).
 
 	``argdescription`` : No command line equivalent
 		Description for help message of the command line argument parser.
@@ -154,8 +139,14 @@ class Job(object):
 		Maximum allowed runtime for the job (as the number of seconds). If the job
 		runs longer than that it will kill itself.
 
+	``fork`` : :option:`--fork`
+		Forks the process and does the work in the child process. The parent
+		process is responsible for monitoring the maximum runtime (this is the
+		default). In non-forking mode the single process does both the work and
+		the runtime monitoring.
+
 	``noisykills`` : :option:`--noisykills`
-		Should an exception be raised when the maximum runtime is exceeded?
+		Should a message be printed when the maximum runtime is exceeded?
 
 	``logfilename`` : :option:`--logfilename`
 		Path/name of the logfile for this job as an UL4 template. Variables
@@ -180,18 +171,11 @@ class Job(object):
 		the same directory as the current logfile that's more than
 		``keepfilelogs`` days old) will be removed at the end of the job.
 
-	``inputencoding`` : :option:`--inputencoding`
-		The encoding to be used for data that is supposed to be unicode, but isn't
-		(e.g. host/user/network info, lines passed to ``self.log`` etc.)
-
-	``inputerrors`` : :option:`--inputerrors`
-		Decoding error handler name (goes with ``inputencoding``)
-
-	``outputencoding`` : :option:`--outputencoding`
+	``encoding`` : :option:`--encoding`
 		The encoding to be used for the logfile.
 
-	``outputerrors`` : :option:`--outputerrors`
-		Encoding error handler name (goes with ``outputencoding``)
+	``errors`` : :option:`--errors`
+		Encoding error handler name (goes with ``encoding``)
 
 	Command line arguments take precedence over instance attributes (if
 	:func:`executewithargs` is used) and those take precedence over class
@@ -205,31 +189,30 @@ class Job(object):
 
 	maxtime = 5 * 60
 
+	fork = True
+
 	noisykills = False
 
-	logfilename = u"~/ll.sisyphus/<?print projectname?>/<?print jobname?>/<?print starttime.format('%Y-%m-%d-%H-%M-%S-%f')?>.sisyphuslog"
-	loglinkname = u"~/ll.sisyphus/<?print projectname?>/<?print jobname?>/current.sisyphuslog"
+	logfilename = "~/ll.sisyphus/<?print projectname?>/<?print jobname?>/<?print format(starttime, '%Y-%m-%d-%H-%M-%S-%f')?>.sisyphuslog"
+	loglinkname = "~/ll.sisyphus/<?print projectname?>/<?print jobname?>/current.sisyphuslog"
 
 	log2file = True
 	log2stdout = False
 	log2stderr = False
 
-	formatlogline = u"[<?print time?>]=[t+<?print time-starttime?>]<?for tag in tags?>[<?print tag?>]<?end for?>: <?print line?>"
+	formatlogline = "[<?print time?>]=[t+<?print time-starttime?>]<?for tag in tags?>[<?print tag?>]<?end for?>: <?print line?>"
 
 	keepfilelogs = 30
 
-	inputencoding = u"utf-8"
-	inputerrors = u"literaldecode"
-
-	outputencoding = u"utf-8"
-	outputerrors = u"strict"
+	encoding = "utf-8"
+	errors = "strict"
 
 	def execute(self):
 		"""
 		Execute the job once. The return value is a one line summary of what the
 		job did. Overwrite in subclasses.
 		"""
-		return u"done"
+		return "done"
 
 	def failed(self):
 		"""
@@ -243,19 +226,18 @@ class Job(object):
 		Return an :mod:`argparse` parser for parsing the command line arguments.
 		This can be overwritten in subclasses to add more arguments.
 		"""
-		p = argparse.ArgumentParser(description=self.argdescription)
-		p.add_argument("-p", "--projectname", dest="projectname", metavar="NAME", help="The name of the project this job belongs to", type=self._string, default=self.projectname)
-		p.add_argument("-j", "--jobname", dest="jobname", metavar="NAME", help="The name of the job", type=self._string, default=self.jobname if self.jobname is not None else self.__class__.__name__)
-		p.add_argument("-m", "--maxtime", dest="maxtime", metavar="SECONDS", help="Maximum number of seconds the job is allowed to run", type=int, default=self.maxtime)
-		p.add_argument("-f", "--log2file", dest="log2file", metavar="FLAG", help="Should the job log into a file?", type=misc.flag, default=self.log2file)
-		p.add_argument("-o", "--log2stdout", dest="log2stdout", metavar="FLAG", help="Should the job log to stdout?", type=misc.flag, default=self.log2stdout)
-		p.add_argument("-e", "--log2stderr", dest="log2stderr", metavar="FLAG", help="Should the job log to stderr?", type=misc.flag, default=self.log2stderr)
-		p.add_argument(      "--keepfilelogs", dest="keepfilelogs", metavar="DAYS", help="Number of days log files are kept", type=float, default=self.keepfilelogs)
-		p.add_argument(      "--inputencoding", dest="inputencoding", metavar="ENCODING", help="Encoding for system data (i.e. crontab etc.)", default=self.inputencoding)
-		p.add_argument(      "--inputerrors", dest="inputerrors", metavar="METHOD", help="Error handling method for encoding errors in system data", default=self.inputerrors)
-		p.add_argument(      "--outputencoding", dest="outputencoding", metavar="ENCODING", help="Encoding for the log file", default=self.outputencoding)
-		p.add_argument(      "--outputerrors", dest="outputerrors", metavar="METHOD", help="Error handling method for encoding errors in log texts", default=self.outputerrors)
-		p.add_argument(      "--noisykills", dest="noisykills", metavar="FLAG", help="Should an exception be raised if the maximum runtime is exceeded?", type=misc.flag, default=self.noisykills)
+		p = argparse.ArgumentParser(description=self.argdescription, epilog="For more info see http://www.livinglogic.de/Python/sisyphus/")
+		p.add_argument("-p", "--projectname", dest="projectname", metavar="NAME", help="The name of the project this job belongs to (default: %(default)s)", default=self.projectname)
+		p.add_argument("-j", "--jobname", dest="jobname", metavar="NAME", help="The name of the job (default: %(default)s)", default=self.jobname if self.jobname is not None else self.__class__.__name__)
+		p.add_argument("-m", "--maxtime", dest="maxtime", metavar="SECONDS", help="Maximum number of seconds the job is allowed to run (default: %(default)s)", type=int, default=self.maxtime)
+		p.add_argument(      "--fork", dest="fork", help="Fork the process and do the work in the child process? (default: %(default)s)", action=misc.FlagAction, default=self.fork)
+		p.add_argument("-f", "--log2file", dest="log2file", help="Should the job log into a file? (default: %(default)s)", action=misc.FlagAction, default=self.log2file)
+		p.add_argument("-o", "--log2stdout", dest="log2stdout", help="Should the job log to stdout? (default: %(default)s)", action=misc.FlagAction, default=self.log2stdout)
+		p.add_argument("-e", "--log2stderr", dest="log2stderr", help="Should the job log to stderr? (default: %(default)s)", action=misc.FlagAction, default=self.log2stderr)
+		p.add_argument(      "--keepfilelogs", dest="keepfilelogs", metavar="DAYS", help="Number of days log files are kept (default: %(default)s)", type=float, default=self.keepfilelogs)
+		p.add_argument(      "--encoding", dest="encoding", metavar="ENCODING", help="Encoding for the log file (default: %(default)s)", default=self.encoding)
+		p.add_argument(      "--errors", dest="errors", metavar="METHOD", help="Error handling method for encoding errors in log texts (default: %(default)s)", default=self.errors)
+		p.add_argument(      "--noisykills", dest="noisykills", help="Should a message be printed if the maximum runtime is exceeded? (default: %(default)s)", action=misc.FlagAction, default=self.noisykills)
 		return p
 
 	def parseargs(self, args=None):
@@ -269,35 +251,53 @@ class Job(object):
 		self.projectname = args.projectname
 		self.jobname = args.jobname
 		self.maxtime = args.maxtime
+		self.fork = args.fork
 		self.noisykills = args.noisykills
 		self.log2file = args.log2file
 		self.log2stdout = args.log2stdout
 		self.log2stderr = args.log2stderr
 		self.keepfilelogs = datetime.timedelta(days=args.keepfilelogs)
-		self.inputencoding = args.inputencoding
-		self.inputerrors = args.inputerrors
-		self.outputencoding = args.outputencoding
-		self.outputerrors = args.outputerrors
+		self.encoding = args.encoding
+		self.errors = args.errors
 		return args
 
+	def _alarm_fork(self, signum, frame):
+		os.kill(self.killpid, signal.SIGTERM) # Kill our child
+		maxtime = datetime.timedelta(seconds=self.maxtime)
+		if self._logfile is not None:
+			self.log.sisyphus.result.kill("Terminated child after {}".format(maxtime))
+			self._logfile.close()
+		if self.noisykills:
+			print("Terminated forked job {} (pid {}) after {}".format(misc.sysinfo.scriptname, misc.sysinfo.pid, maxtime))
+		os._exit(1)
+
+	def _alarm_nofork(self, signum, frame):
+		self._prefix = ""
+		maxtime = datetime.timedelta(seconds=self.maxtime)
+		if self._logfile is not None:
+			self.log.sisyphus.result.kill("Terminated after {}".format(maxtime))
+			self._logfile.close()
+		if self.noisykills:
+			print("Terminated job {} (pid {}) after {}".format(misc.sysinfo.scriptname, misc.sysinfo.pid, maxtime))
+		os._exit(1)
+
 	def _handleexecution(self):
 		"""
 		Handle executing the job including handling of duplicate or hanging jobs.
 		"""
 		self.info = AttrDict()
-		self.info.sysinfo = misc.SysInfo(self.inputencoding, self.inputerrors)
-		self.info.projectname = self._string(self.projectname)
-		self.info.jobname = self._string(self.jobname)
+		self.info.sysinfo = misc.sysinfo
+		self.info.projectname = self.projectname
+		self.info.jobname = self.jobname
 		self.info.maxtime = self.maxtime
 		self._prefix = ""
-		maxtimedelta = datetime.timedelta(seconds=self.maxtime)
 
 		# Obtain a lock on the script file to make sure we're the only one running
-		with open(self.info.sysinfo.scriptname, "rb") as f:
+		with open(misc.sysinfo.scriptname, "rb") as f:
 			try:
 				fcntl.flock(f, fcntl.LOCK_EX | fcntl.LOCK_NB)
-			except IOError, exc:
-				if exc[0] not in (errno.EACCES, errno.EAGAIN): # some other error
+			except IOError as exc:
+				if exc.errno not in (errno.EACCES, errno.EAGAIN): # some other error
 					raise
 				# The previous invocation of the job is still running
 				return # Return without calling :meth:`execute`
@@ -309,48 +309,46 @@ class Job(object):
 			self._getcrontab() # Get crontab
 			self.lineno = 1 # Current line number
 			self.log = Tag(self._log) # Create tagged logger
-			self._formatlogline = ul4c.compile(self.formatlogline.replace("\n", "").replace("\r", "") + u"\n") # Log line formatting template
+			self._formatlogline = ul4c.Template(self.formatlogline.replace("\n", "").replace("\r", "") + "\n", "formatlogline") # Log line formatting template
 			self._createlog() # Create log file and link
 
-			self.log.sisyphus.init(u"{} (max run time {}; parent pid {})".format(self.info.sysinfo.scriptname, maxtimedelta, self.info.sysinfo.pid))
+			self.log.sisyphus.init("{} (max time {}; pid {})".format(misc.sysinfo.scriptname, datetime.timedelta(seconds=self.maxtime), misc.sysinfo.pid))
 
-			# Fork the process; the child will do the work; the parent will monitor the maximum runtime
-			pid = os.fork()
-
-			if pid: # We are the parent process
-				# set a signal that kills the child process after the maximum runtime
-				def alarm(signum, frame):
-					os.kill(pid, signal.SIGTERM) # Kill our child
-					self.log.sisyphus.error.kill(u"Terminated after {}".format(maxtimedelta))
-					if self.noisykills:
-						raise MaximumRuntimeExceeded(self.maxtime)
-					sys.exit(1)
-				signal.signal(signal.SIGALRM, alarm)
+			if self.fork: # Forking mode?
+				# Fork the process; the child will do the work; the parent will monitor the maximum runtime
+				self.killpid = pid = os.fork()
+				if pid: # We are the parent process
+					# set a signal to kill the child process after the maximum runtime
+					signal.signal(signal.SIGALRM, self._alarm_fork)
+					signal.alarm(self.maxtime)
+					os.waitpid(pid, 0) # Wait for the child process to terminate
+					return # Exit normally
+				self.log.sisyphus.init("forked worker child (child pid {})".format(os.getpid()))
+			else: # We didn't fork
+				# set a signal to kill ourselves after the maximum runtime
+				signal.signal(signal.SIGALRM, self._alarm_nofork)
 				signal.alarm(self.maxtime)
-				os.wait() # Wait for the child process to terminate
-				return # Exit normally
-			# From now on we are in the child process
-			self.log.sisyphus.init(u"forked worker child (child pid {})".format(os.getpid()))
 
 			try:
 				with url.Context():
 					result = self.execute()
 				self._cleanupoldlogs() # Clean up old logfiles
-			except BaseException, exc:
+			except BaseException as exc:
 				# log the error to the logfile, because the job probably didn't have a chance to do it
 				self.log.sisyphus.exc(exc)
-				result = u"failed with {}".format(self._exc(exc))
+				result = "failed with {}".format(self._exc(exc))
 				self.log.sisyphus.result.fail(result)
 				self.failed()
 				raise
 			else:
 				# log the result
-				self.log.sisyphus.result.ok(self._string(result))
+				self.log.sisyphus.result.ok(result)
 			finally:
 				if self._logfile is not None:
 					self._logfile.close()
 				fcntl.flock(f, fcntl.LOCK_UN | fcntl.LOCK_NB)
-			os._exit(0)
+			if self.fork:
+				os._exit(0)
 
 	@contextlib.contextmanager
 	def prefix(self, prefix):
@@ -374,20 +372,18 @@ class Job(object):
 		if self.log2file or self.log2stdout or self.log2stderr:
 			timestamp = datetime.datetime.now()
 			for text in texts:
-				text = self._string(text)
 				if isinstance(text, BaseException):
 					if "exc" not in tags:
 						tags += ("exc",)
-					tb = u"".join(traceback.format_tb(sys.exc_info()[-1]))
-					text = u"{tb}\n{exc}".format(tb=tb, exc=self._exc(text))
-				elif not isinstance(text, unicode):
-					text = self._string(pprint.pformat(text))
+					tb = "".join(traceback.format_tb(sys.exc_info()[-1]))
+					text = "{tb}\n{exc}".format(tb=tb, exc=self._exc(text))
+				elif not isinstance(text, str):
+					text = pprint.pformat(text)
 				lines = text.splitlines()
 				if lines and not lines[-1].strip():
 					del lines[-1]
 				for line in lines:
 					text = self._formatlogline.renders(line=self._prefix+line, time=timestamp, tags=tags, **self.info)
-					text = text.encode(self.outputencoding, self.outputerrors)
 					if self.log2file:
 						self._logfile.write(text)
 						self._logfile.flush()
@@ -403,20 +399,11 @@ class Job(object):
 		"""
 		Reads the source code of the script into ``self.source``.
 		"""
+		scriptname = misc.sysinfo.scriptname.rstrip("c")
 		try:
-			with open(self.info.sysinfo.scriptname.rstrip("c"), "rb") as f:
-				source = f.read()
-				lines = source.splitlines()
-				# find encoding in the first two lines
-				encoding = self.inputencoding
-				if lines and lines[0].startswith(codecs.BOM_UTF8):
-					encoding = "utf-8"
-				else:
-					for line in lines[:2]:
-						match = encodingdeclaration.search(line)
-						if match is not None:
-							encoding = match.group(1)
-				self.source = source.decode(encoding, self.inputerrors)
+			encoding = tokenize.detect_encoding(open(scriptname, "rb").readline)[0]
+			with open(scriptname, "r", encoding=encoding, errors="replace") as f:
+				self.source = f.read()
 		except IOError: # Script might have called ``os.chdir()`` before
 			self.source = None
 
@@ -424,7 +411,8 @@ class Job(object):
 		"""
 		Reads the current crontab into ``self.crontab``.
 		"""
-		self.crontab = self._string(os.popen("crontab -l 2>/dev/null").read())
+		with os.popen("crontab -l 2>/dev/null") as f:
+			self.crontab = f.read()
 
 	def _createlog(self):
 		"""
@@ -435,18 +423,18 @@ class Job(object):
 		self._loglinkname = None
 		if self.log2file:
 			# Create the log file
-			logfilename = ul4c.compile(self.logfilename).renders(**self.info)
+			logfilename = ul4c.Template(self.logfilename, "logfilename").renders(**self.info)
 			lf = self._logfilename = url.File(logfilename).abs()
-			self._logfile = lf.openwrite()
+			self._logfile = lf.open(mode="w", encoding=self.encoding, errors=self.errors)
 			if self.loglinkname is not None:
 				# Create the log link
-				loglinkname = ul4c.compile(self.loglinkname).renders(**self.info)
+				loglinkname = ul4c.Template(self.loglinkname, "loglinkname").renders(**self.info)
 				ll = self._loglinkname = url.File(loglinkname).abs()
 				lf = self._logfilename
 				try:
 					lf.symlink(ll)
-				except OSError, exc:
-					if exc[0] == errno.EEXIST:
+				except OSError as exc:
+					if exc.errno == errno.EEXIST:
 						ll.remove()
 						lf.symlink(ll)
 					else:
@@ -476,23 +464,19 @@ class Job(object):
 					self.log.sisyphus.info("Removing logfile {}".format(fileurl.local()))
 					fileurl.remove()
 
-	def _string(self, s):
-		"""
-		Convert :var:`s` to unicode if it's a :class:`str`.
-		"""
-		if isinstance(s, str):
-			s = s.decode(self.inputencoding, self.inputerrors)
-		return s
-
 	def _exc(self, exc):
 		"""
 		Format an exception object for logging.
 		"""
 		if exc.__class__.__module__ not in ("__builtin__", "exceptions"):
-			fmt = u"{0.__class__.__module__}.{0.__class__.__name__}: {0}"
+			fmt = "{0.__class__.__module__}.{0.__class__.__name__}: {1}"
 		else:
-			fmt = u"{0.__class__.__name__}: {0}"
-		return fmt.format(exc)
+			fmt = "{0.__class__.__name__}: {1}"
+		try:
+			strexc = str(exc)
+		except UnicodeError:
+			strexc = "?"
+		return fmt.format(exc, strexc)
 
 
 class Tag(object):
